# -*- encoding: utf-8 -*-
from __future__ import absolute_import
__version__ = '0.1'


from .explanation import HumanExplanation

from .features import apply_rules_layer
from .features import FeatureWeightGroup

from .scoring import score

from .formatters import (
	percent,
	delta_percent,
	value,
	text,
	integer,
	value_simplified
)



# -*- encoding: utf-8 -*-

import copy

from jinja2 import Environment, PackageLoader, select_autoescape
from eli5.formatters.html import format_hsl, weight_color_hsl

from .helpers import (
	_extract_from_dictionary,
	_extract_formatted_value,
	_extract_label
)
from .features import apply_rules_layer
from six import iteritems

env = Environment(
	loader=PackageLoader('elih', 'templates'),
	autoescape=select_autoescape(['html'])
)

env.filters.update(dict(
	weight_color=lambda w, w_range: format_hsl(weight_color_hsl(w, w_range))
))


def translate_explanation(explanation, dictionary):
	'''A simple method that returns an ELI5 Explanation object with features renamed
	following the given dictionary.
	'''
	labels_dictionary = _extract_from_dictionary(dictionary)
	translated_explanation = copy.deepcopy(explanation)
	for feature_weight in translated_explanation.targets[0].feature_weights.pos + translated_explanation.targets[0].feature_weights.neg:
		if feature_weight.feature in labels_dictionary:
			feature_weight.feature = labels_dictionary[feature_weight.feature]

	return translated_explanation


def translate_keys(dict, dictionary):
	'''A simple method to translate dict keys to new labels using a dictionary.
	'''
	new_dict = {}
	labels_dictionary = _extract_from_dictionary(dictionary)
	for k, v in iteritems(dict):
		if k in labels_dictionary:
			new_dict[labels_dictionary[k]] = v
		else:
			new_dict[k] = v
	return new_dict


def apply_interpretors(interpretors, all_variables_with_value, all_variables_with_formatted_value):
	interpretations = {}
	for interpretation_code, interpretation_rules in iteritems(interpretors):
		if interpretation_rules['assert'](all_variables_with_value):
			interpretations[interpretation_code] = interpretation_rules['interpretation'](all_variables_with_formatted_value)
		else:
			if 'not_interpretation' in interpretation_rules:
				interpretations[interpretation_code] = interpretation_rules['not_interpretation'](all_variables_with_formatted_value)
	return interpretations


class HumanExplanation(object):
	'''A layer on top of ELI5 Explanation object to provide additional services
	'''

	def __init__(
			self,
			explanation,
			rules_layers,
			additional_features=None,
			dictionary=None,
			scoring=None,
			interpretors={}
	):
		if type(rules_layers) is dict:
			# Only one layer of rules was provided
			rules_layers = [rules_layers]
		if additional_features is None:
			self.additional_features = {}
		else:
			self.additional_features = self._translate_additional_features(additional_features, dictionary)

		if dictionary is None:
			self.dictionary = {}
		else:
			self.dictionary = dictionary

		self.scoring = scoring

		# Apply rules layers
		self.rules_layers = rules_layers
		self.explanation_layers = []
		for index, rules in enumerate(rules_layers):
			if index > 0:
				previous_explanation = self.explanation_layers[index - 1]
			else:
				previous_explanation = explanation
			new_explanation = apply_rules_layer(previous_explanation, rules, additional_features, dictionary, scoring)
			self.explanation_layers.append(new_explanation)

		# Aggregate (valued) variables coming from everywhere, then apply interpretors
		all_variables_with_value = {}  # for asserts
		all_variables_with_formatted_value = {}  # for interpretations
		for layer in self.explanation_layers:
			for feature in layer.targets[0].feature_weights.pos + layer.targets[0].feature_weights.neg:
				if feature.value is not None:
					all_variables_with_value[feature.feature] = feature.value
					if feature.formatted_value is not None:
						all_variables_with_formatted_value[feature.feature] = {
							'value': feature.value,
							'formatted_value': feature.formatted_value
						}

		for variable, value in iteritems(self.additional_features):
			if 'value' in value:
				all_variables_with_value[variable] = value['value']
				if 'formatted_value' in value:
					all_variables_with_formatted_value[variable] = {
						'value': value['value'],
						'formatted_value': value['formatted_value']
					}
				else:
					all_variables_with_formatted_value[variable] = {
						'value': value['value']
					}

		self.interpretations = apply_interpretors(interpretors, all_variables_with_value, all_variables_with_formatted_value)

	def _translate_additional_features(self, additional_features, dictionary):
		new_dict = {}
		for feature_name, value in iteritems(additional_features):
			feature_dict = {}
			# value is not supposed to change because it's coming directly from the source (so, can be controlled)
			feature_dict['value'] = value
			# formatted_value takes the value and formats it
			feature_dict['formatted_value'] = _extract_formatted_value(value, dictionary, feature_name)
			feature_dict['label'] = _extract_label(dictionary, feature_name)

			new_dict[feature_name] = feature_dict
		return new_dict

	def __repr__(self):
		all_repr = ''
		for layer in self.explanation_layers:
			all_repr += '{}(explanation={}, additional_features={}, interpretations={})'.format(
				'HumanExplanation',
				layer.__repr__(),
				translate_keys(self.additional_features, self.dictionary),
				self.interpretations
			)
		return all_repr

	def _repr_html_(self):
		template = env.get_template('explanation.html')
		layers = []
		for index, layer in enumerate(self.explanation_layers):
			features = layer.targets[0].feature_weights.pos + layer.targets[0].feature_weights.neg
			weight_range = abs(max([f.weight for f in features]))
			layers.append({
				'features': features,
				'weight_range': weight_range
			})
		return template.render(
			layers=layers,
			additional_features=self.additional_features,
			interpretations=self.interpretations
		)

	def to_dict(self):
		return_obj = {}

		# Explanation layers
		layers = []
		for layer in self.explanation_layers:
			layer_obj = {
				"pos": [f.to_dict() for f in layer.targets[0].feature_weights.pos],
				"neg": [f.to_dict() for f in layer.targets[0].feature_weights.neg]
			}
			layers.append(layer_obj)
		return_obj['explanation_layers'] = layers

		# Additional variables
		return_obj['additional_variables'] = translate_keys(self.additional_features, self.dictionary)

		# Interpretations
		return_obj['interpretations'] = self.interpretations

		return return_obj



# -*- encoding: utf-8 -*-

import copy
import fnmatch

from six import iteritems
from past.builtins import basestring
from eli5.base import FeatureWeights, FeatureWeight

from .helpers import _extract_mapped_value, _extract_formatted_value


class EnrichedFeatureWeight(FeatureWeight):
	"""Enriches ELI5 FeatureWeight with additional 'human explanation' data like
	a human readable feature label, description, format, ...
	"""

	def __init__(self, *args, **kwargs):
		self.dictionary = kwargs['dictionary'] if 'dictionary' in kwargs else None
		self.formatted_value = kwargs['formatted_value'] if 'formatted_value' in kwargs else None
		self.score = kwargs['score'] if 'score' in kwargs else None
		if 'dictionary' in kwargs:
			del kwargs['dictionary']
		if 'formatted_value' in kwargs:
			del kwargs['formatted_value']
		if 'score' in kwargs:
			del kwargs['score']
		FeatureWeight.__init__(self, *args, **kwargs)

	def __repr__(self):
		return "{}(feature='{}', weight={}, score={}, std={}, value={}, formatted_value={}, dictionary={})".format(
			'EnrichedFeatureWeight',
			self.feature,
			self.weight,
			self.score,
			self.std,
			self.value,
			self.formatted_value,
			self.dictionary
		)

	def to_dict(self):
		return {
			"feature": self.feature,
			"weight": self.weight,
			"score": self.score,
			"std": self.std,
			"value": self.value,
			"formatted_value": self.formatted_value,
			"label": self.dictionary["label"] if self.dictionary is not None and 'label' in self.dictionary else None
		}


class FeatureWeightGroup(EnrichedFeatureWeight):
	"""A class derived from ELI5's FeatureWeight to store the initial FeatureWeight-s
	regrouped to build up this new FeatureWeight.
	"""

	def __init__(self, *args, **kwargs):
		self.group = kwargs['group'] if 'group' in kwargs else []
		if 'group' in kwargs:
			del kwargs['group']
		EnrichedFeatureWeight.__init__(self, *args, **kwargs)

	def __repr__(self):
		return "{}(feature='{}', weight={}, score={}, std={}, value={}, formatted_value={}, dictionary={}, group={})".format(
			'FeatureWeightGroup',
			self.feature,
			self.weight,
			self.score,
			self.std,
			self.value,
			self.formatted_value,
			self.dictionary,
			self.group
		)

	def to_dict(self):
		return {
			"feature": self.feature,
			"weight": self.weight,
			"score": self.score,
			"std": self.std,
			"value": self.value,
			"formatted_value": self.formatted_value,
			"label": self.dictionary["label"] if self.dictionary is not None and 'label' in self.dictionary else None,
			"group": [f.to_dict() for f in self.group]
		}


def apply_rules_layer(explanation, rules, additional_features=None, dictionary=None, scoring=None):
	"""Regroups several feature weights into one brand new feature weight, whose weight is the sum of the ones from the underlying feature weights.

	The new feature weight will be created inside the FeatureWeights object of the explanation, while the previous weights are discarded.

	Now, it only works with binary classifiers (meaning on the first TargetExplanation object)

	TODO: add support for multiple targets.

	Args:
		explanation: An ELI5 Explanation object (typically the output of explain_prediction).
		rules: a dictionary of rules whose keys are new grouped feature names and whose values are either:
			- lists of exact feature names that will be regrouped
			- string for pattern matching in a "Unix-file" flavour (e.g. 'Embarked=*', 'Class*')
		additional_features: (optional) a dictionary with additional variables that may be used to map values
		dictionary: (optional) a dictionary that allows mapping values and labels to features

	Returns:
		An Explanation object with the new grouped features

	"""

	# Otherwise the object is modified by reference
	explanation = copy.deepcopy(explanation)

	if additional_features is None:
		additional_features = {}
	if dictionary is None:
		dictionary = {}

	# Start by getting out the generic rules (e.g. 'Variable=*') cause they are processed in a different way:
	generic_rules = {
		v: k for (k, v) in rules.items() if isinstance(v, basestring)
	}

	# Then reverse the other rules:
	# {'A': ['1', '2'], 'B': ['3']} to {'1': 'A', '2': 'A', '3': 'B'}
	new_rules = {old_field: grouped_field for (grouped_field, old_fields) in rules.items() if not isinstance(old_fields, basestring) for old_field in old_fields}

	new_weights = {}
	for feature_weight in explanation.targets[0].feature_weights.pos + explanation.targets[0].feature_weights.neg:

		if feature_weight.feature == '<BIAS>':
			continue

		matched = False

		# 'Old feature' matches a standard rule?
		if feature_weight.feature in new_rules:
			matched = True
			# Feature is grouped with others
			grouped_feature = new_rules[feature_weight.feature]
			if grouped_feature in new_weights:
				# Add weight to already created grouped feature
				new_weights[grouped_feature]['weight'] = new_weights[grouped_feature]['weight'] + feature_weight.weight
			else:
				# Create new grouped feature
				_mapped_value = _extract_mapped_value(additional_features, dictionary, grouped_feature)
				new_weights[grouped_feature] = {
					'feature': grouped_feature,
					'weight': feature_weight.weight,
					'score': scoring(feature_weight.weight) if scoring is not None else None,
					'std': None,
					'value': _mapped_value,
					'formatted_value': _extract_formatted_value(_mapped_value, dictionary, grouped_feature),
					'group': [],
					'dictionary': dictionary[grouped_feature] if grouped_feature in dictionary else {}
				}
			new_weights[grouped_feature]['group'].append({
				'feature': feature_weight.feature,
				'weight': feature_weight.weight,
				'score': scoring(feature_weight.weight) if scoring is not None else None,
				'std': feature_weight.std,
				'value': feature_weight.value,
				'formatted_value': _extract_formatted_value(feature_weight.value, dictionary, feature_weight.feature),
				'dictionary': dictionary[feature_weight.feature] if feature_weight.feature in dictionary else {}
			})

		# Match generic rule?
		for rule, grouped_feature in iteritems(generic_rules):
			if fnmatch.fnmatch(feature_weight.feature, rule):
				matched = True
				if grouped_feature in new_weights:
					# Add weight to already created grouped feature
					new_weights[grouped_feature]['weight'] = new_weights[grouped_feature]['weight'] + feature_weight.weight
				else:
					# Create new grouped feature
					_mapped_value = _extract_mapped_value(additional_features, dictionary, grouped_feature)
					new_weights[grouped_feature] = {
						'feature': grouped_feature,
						'weight': feature_weight.weight,
						'score': scoring(feature_weight.weight) if scoring is not None else None,
						'std': None,
						'value': _mapped_value,
						'formatted_value': _extract_formatted_value(_mapped_value, dictionary, grouped_feature),
						'group': [],
						'dictionary': dictionary[grouped_feature] if grouped_feature in dictionary else {}
					}
				new_weights[grouped_feature]['group'].append({
					'feature': feature_weight.feature,
					'weight': feature_weight.weight,
					'score': scoring(feature_weight.weight) if scoring is not None else None,
					'std': feature_weight.std,
					'value': feature_weight.value,
					'formatted_value': _extract_formatted_value(feature_weight.value, dictionary, feature_weight.feature),
					'dictionary': dictionary[feature_weight.feature] if feature_weight.feature in dictionary else {}
				})

		# No match for this feature => remains the same
		if not matched:
			# Feature remains the same ('not a group')
			new_weights[feature_weight.feature] = {
				'feature': feature_weight.feature,
				'weight': feature_weight.weight,
				'score': scoring(feature_weight.weight) if scoring is not None else None,
				'std': feature_weight.std,
				'value': feature_weight.value,
				'formatted_value': _extract_formatted_value(feature_weight.value, dictionary, feature_weight.feature),
				'dictionary': dictionary[feature_weight.feature] if feature_weight.feature in dictionary else {}
			}

	new_features = []
	for new_feature, new_feature_weight in iteritems(new_weights):
		if 'group' in new_feature_weight:
			obj = FeatureWeightGroup(
				feature=new_feature_weight['feature'],
				weight=new_feature_weight['weight'],
				score=new_feature_weight['score'],
				std=new_feature_weight['std'],
				value=new_feature_weight['value'],
				formatted_value=new_feature_weight['formatted_value'],
				group=[
					EnrichedFeatureWeight(
						feature=f['feature'],
						weight=f['weight'],
						score=f['score'],
						std=f['std'],
						value=f['value'],
						formatted_value=f['formatted_value'],
						dictionary=f['dictionary']
					) for f in new_feature_weight['group']
				],
				dictionary=new_feature_weight['dictionary']
			)
		else:
			obj = EnrichedFeatureWeight(
				feature=new_feature_weight['feature'],
				weight=new_feature_weight['weight'],
				score=new_feature_weight['score'],
				std=new_feature_weight['std'],
				value=new_feature_weight['value'],
				formatted_value=new_feature_weight['formatted_value'],
				dictionary=new_feature_weight['dictionary']
			)
		new_features.append(obj)

	# To build a FeatureWeights object, we then need to sort features by weight and separate positives and negatives:
	feature_weights = FeatureWeights(
		pos=sorted([f for f in new_features if f.weight >= 0], key=(lambda o: abs(o.weight)), reverse=True),
		neg=sorted([f for f in new_features if f.weight < 0], key=(lambda o: abs(o.weight)), reverse=False)
	)

	explanation.targets[0].feature_weights = feature_weights

	return explanation



# -*- encoding: utf-8 -*-

from six import iteritems
from past.builtins import basestring


def _extract_from_dictionary(dictionary, field='label'):
	'''An ELIH dictionary can embed, for each variable, either just a 'human label' or more than that
	(description, format, etc.). This helper extracts the labels dictionary from it.
	'''
	extracted_labels = {}
	for k, v in iteritems(dictionary):
		if isinstance(v, basestring):
			extracted_labels[k] = v
		elif type(v) is dict and field in v:
			extracted_labels[k] = v[field]
	return extracted_labels


def _extract_mapped_value(additional_features, dictionary, grouped_feature):
	value_from = None
	if grouped_feature in dictionary and 'value_from' in dictionary[grouped_feature]:
		try:
			value_from = additional_features[dictionary[grouped_feature]['value_from']]
		except IndexError:
			raise IndexError("Variable {} required but not provided with the additional features.".format(dictionary[grouped_feature]['value_from']))
	return value_from


def _extract_label(dictionary, feature_name):
	label = None
	if feature_name in dictionary and 'label' in dictionary[feature_name]:
		try:
			label = dictionary[feature_name]['label']
		except:
			raise Exception("Exception when generating the label for variable {}.".format(feature_name))
	return label


def _extract_formatted_value(value, dictionary, feature_name):
	formatted_value = None
	if value is None:
		return None
	if feature_name in dictionary and 'formatter' in dictionary[feature_name]:
		try:
			formatted_value = dictionary[feature_name]['formatter'](value)
		except:
			raise Exception("Exception when generating the formatted value for variable {} with value {}.".format(feature_name, value))
	return formatted_value



# -*- encoding: utf-8 -*-

from setuptools import setup

setup(
	name="elih",
	version="0.2",
	description="A library to translate Machine Learning classifiers predictions in a human understandable and simplified form.",
	url="https://github.com/fvinas/elih",
	download_url="https://github.com/fvinas/elih/archive/0.2.tar.gz",
	author="Fabien Vinas",
	author_email="fabien.vinas@gmail.com",
	license="",
	packages=["elih"],
	install_requires=[
		'eli5 >= 0.6.1',
		'six',
		'future'
	],
	zip_safe=False
)



# -*- encoding: utf-8 -*-


def mapper(dictionary):
	"""A formatter that map a value to another given a dict.
	Useful to quickly map categorical variables with very few levels (e.g a "sex" variable)
	to something more understandable (or translated).
	"""
	return (lambda a: dictionary[a] if a in dictionary else a)


def percent(decimals=1):
	"""A formatter to display a float as a percentage with a given number of decimals.

	Args:
		decimals: The number of decimals to display.

	Returns:
		A formatter function f returning f(a), being variable 'a' displayed as a percentage.
		Please note the formatter function will multiply the float by 100 to get a percentage.

	"""
	return (lambda a: "{:+.{prec}f}%".format(100.0 * a, prec=decimals))


def delta_percent(decimals=1):
	"""A formatter to display the float delta as a percentage with a given number of decimals.
	It's different from the percent formatter in that it's prefixed with a sign (+ or -).

	Args:
		decimals: The number of decimals to display.

	Returns:
		A formatter function f returning f(a), being variable 'a' displayed as a signed percentage.
		Please note the formatter function will multiply the float by 100 to get a percentage.

	"""
	return (lambda a: "{:+.{prec}f}%".format(100.0 * a, prec=decimals))


def value(decimals=1, unit="", sign=""):
	"""A formatter to display a float with a given unit.

	Args:
		decimals = 1: The number of decimals to display.
		unit = "": The unit symbol to display after the delta value.
		sign = "": Allows to move to a signed version ('+5.3M€' vs '5.3M€'). Use sign='+' for this.

	Returns:
		A formatter function f returning f(a), being variable 'a' accompanied with a unit.

	"""
	return (lambda a: "{:{sign},.{prec}f} {unit}".format(1.0 * a, sign=sign, unit=unit, prec=decimals))


def text():
	"""A simple formatter to display a text as it is.

	Returns:
		A formatter function f returning f(a) being variable 'a' displayed as a text.
	"""
	return (lambda a: "{}".format(a))


def integer():
	"""A simple formatter to display an integer with no unit.

	Returns:
		A formatter function f returning f(a) being variable 'a' displayed as an integer.
	"""
	return value(decimals=0, unit="")


def value_simplified(decimals=1, unit="", prefixes=['k', 'M', 'B'], sign=""):
	"""A formatter to display the given float with a given unit in a simplified way (e.g. 10k for 10.000)

	Args:
		decimals = 1: The number of decimals to display.
		unit = "": The unit symbol to concatenate to the value.
		prefixes = ['k', 'M', 'B']: The simplifier symbols to use for 1.000s, 1.000.000s and 1.000.000.000s.
		sign = "": Allows to move to a signed version ('+5.3M€' vs '5.3M€'). Use sign='+' for this.

	Returns:
		A formatter function f returning f(a), being variable 'a' simplified and accompanied with a unit.

	"""
	def formatter(a):
		prefix = ""
		if a >= 1000000000:
			a = a * 1.0 / 1000000000
			prefix = prefixes[2]
		elif a >= 1000000:
			a = a * 1.0 / 1000000
			prefix = prefixes[1]
		elif a >= 1000:
			a = a * 1.0 / 1000
			prefix = prefixes[0]
		return "{:{sign},.{decimals}f} {prefix}{unit}".format(1.0 * a, sign=sign, decimals=decimals, prefix=prefix, unit=unit)
	return formatter
